/*****************************************************************************************
FUNCTIONALITY :
            Creating lib environment and checking it against original environment for discrepancies.
            This macro should be executed on the fresh new session.


USAGE :

            %sas_validate_schema (version= ,  reportpath=.);

PARAMETERS:
            version     :  Any One of -  M1MI42, F1MI42, M2MI42, M1TKMI43, M2TKMI43, F2MITK43

            reportpath  :  Provide path for reports.

OUTPUT :
            Script generate 3 reports as output at default files location.
            1. schema_diff_di_data.txt
            2. schema_diff_didm_stg.txt
            3. schema_diff_di_mon.txt

HISTORY:    25MAY2012  sinavk : Created macro.

*****************************************************************************************/

%macro sas_validate_schema(version    = ,
                           reportpath =.
);

   %let version= %upcase(&version);
   options lrecl=max notes;

   %local lib_path reqlib;

   %if &version=F1MI42   or
       &version=D2TKMI42 or
       &version=M1MI42   or
       &version=M2MI42   or
       &version=M1TKMI43 or
       &version=M2TKMI43 or
       &version=F2TKMI43 %then %do;

      %local x;
      %let x = %di_util_misc_path();

      filename f "&x./dbmsc/metadata_4x.cport";

      proc cimport library=work infile=f extendsn=no; quit;

      filename f clear;

      %compare_sas_lib(baseline=ref_dm,compare=di_data,reportpath=&reportpath, include_fmt_in_diff=1, version=&version);

      %compare_sas_lib(baseline=ref_stg,compare=didm_stg,reportpath=&reportpath, include_fmt_in_diff=1, version=&version);

      %compare_sas_lib(baseline=ref_mon,compare=di_mon,reportpath=&reportpath, include_fmt_in_diff=1,version=&version);

      proc datasets lib=work kill nolist;
      quit;

   %end;
   %else %do;
      %put ERROR : Incorrect version ;
   %end;

%mend sas_validate_schema;

/*********************************************************************************************************************/
/*********************************************************************************************************************/

%macro compare_sas_lib(baseline            =,
                       compare             =,
                       reportpath          =,
                       include_fmt_in_diff =,
                       version             =
);

   option nolabel;

   proc datasets lib=work nolist;
      save sas_base_tab_columns sas_base_ind_columns sas_base_constraints / memtype=data;
   quit;

   %local version1;
   %let version1 = &VERSION ;

   %if "&version" = "F2TKMI43" %then %do;
      %let version1 = M2TKMI43 ;
   %end;

   /* Columns for reference version */
   proc sql;
      create table work.base_tab_columns as
      SELECT *
        FROM work.sas_base_tab_columns
       WHERE LIBNAME="%UPCASE(&baseline)"
         and version="&version1"
       order by memname,name;
   quit;

   /* Indexes for reference version */

   proc sql;
      create table work.base_ind_columns as
      SELECT *
        FROM work.sas_base_ind_columns
       WHERE LIBNAME="%UPCASE(&baseline)"
         and version="&version1"
       order by memname,indxname,indxpos;
   quit;

   /* Constraints for reference version */

   proc sql;
      create table work.base_constraints as
      select *
        FROM work.sas_base_constraints
       where library ="%UPCASE(&baseline)"
         and version="&version1"
     /*  order by memname */  ;
   quit;


   /* Primary Key for reference version */

   data work.base_pk_columns;
      set work.base_constraints(keep  = memname constraint_name column_name  constraint_type
                                where = (constraint_type ='PRIMARY'));
      by memname;
      attrib COLUMN_NAMES length=$300;
      retain COLUMN_NAMES '';
      if first.memname then
         COLUMN_NAMES = column_name;
      else
         COLUMN_NAMES=catx(" ",COLUMN_NAMES,column_name);
      IF last.memname then output;
      drop constraint_type column_name;
   run;

   /* Not Null for reference version */

   data work.base_not_null_cons;
      set work.base_constraints(keep  = memname constraint_name column_name  constraint_type
                                where = (constraint_type ='NOT NULL'));
      by memname;
      drop constraint_name ;
   run;

   proc sort data=work.base_not_null_cons;
      by memname column_name;
   run;

   /* Make Indexes metadata readable */

   data work.base_index_columns(keep=libname memname name indxname idxusage  unique_index icols);
      set work.base_ind_columns;
      by libname memname indxname idxusage indxpos;
      length icols $500;
      length unique_index $100;
      retain icols;
      if Upcase(unique)='YES' then unique_index='Unique';
      if Upcase(unique)='NO' then  unique_index='NotUnique';
      icols=UPCASE(catx(', ',icols,name));

      if last.memname or
         last.indxname or
         last.idxusage then do;

         indxname =UPCASE(indxname );
         output;
         icols='';
      end;
   run;

   proc sort data=work.base_index_columns;
      by memname indxname;
   run;

   proc sql;
      create table work.comp_tab_columns as
      SELECT *
        FROM DICTIONARY.COLUMNS
       WHERE LIBNAME="%UPCASE(&compare)"
       order by memname,name;
   quit;

   /* Indexes for compare version */

   proc sql;
      create table work.comp_ind_columns as
      SELECT *
        FROM sashelp.vindex
       WHERE LIBNAME="%UPCASE(&compare)"
       order by memname,indxname,indxpos;
   quit;

   /* Constraints for compare version */

   data dd_constraint_column_usage;
   retain column_order;
   set sashelp.Vcncolu(where=(table_catalog ="%UPCASE(&compare)"));
   by table_name;
   if first.table_name then column_order = 1;
   else column_order = column_order+1;
   run;

   proc sql;
      create table work.comp_constraints as
      select cc.table_catalog as library,
             cc.table_schema  as libschema,
             cc.table_name    as memname,
             cc.column_name,
             cc.constraint_name,
             tc.constraint_type,
			 cc.column_order
        FROM dd_constraint_column_usage as cc
        join dictionary.table_constraints as tc
          on (
                cc.table_catalog   = tc.table_catalog   and
                cc.table_schema    = tc.table_schema    and
                cc.table_name      = tc.table_name      and
                cc.constraint_name = tc.constraint_name
             )
       where cc.table_catalog ="%UPCASE(&compare)"
        order by cc.table_name,  cc.column_order ;
   quit;

   /* Primary Key for compare version */

   data work.comp_pk_columns;
      set work.comp_constraints(keep  = memname constraint_name column_name  constraint_type
                                where = (constraint_type ='PRIMARY'));
      by memname;
      attrib COLUMN_NAMES length=$300;
      retain COLUMN_NAMES '';
      if first.memname then
         COLUMN_NAMES = column_name;
      else
         COLUMN_NAMES=catx(" ",COLUMN_NAMES,column_name);
      IF last.memname then output;
      drop constraint_type column_name;
   run;

   /* Not Null for compare version */

   data work.comp_not_null_cons;
      set work.comp_constraints(keep  = memname constraint_name column_name  constraint_type
                                where = (constraint_type ='NOT NULL'));
      by memname;
      drop constraint_name ;
   run;

   proc sort data=work.comp_not_null_cons;
      by memname column_name;
   run;

   /* Indexes for compare version */

   data work.comp_index_columns(keep=libname memname name indxname idxusage unique_index icols);
      set work.comp_ind_columns;
      by libname memname indxname idxusage indxpos;
      length icols $500;
      retain icols;
      length unique_index $100;
      if Upcase(unique)='YES' then
         unique_index='Unique';
      if Upcase(unique)='NO' then
         unique_index='NotUnique';

      icols=UPCASE(catx(', ',icols,name));

      if last.memname  or
         last.indxname or
         last.idxusage then do;

         indxname =UPCASE(indxname );
         output;
         icols='';
      end;
   run;

   proc sort data=work.comp_index_columns;
      by memname indxname;
   run;

   *Tables New in Compare ;

   data work.new_tables_in_comp work.tables_not_in_comp work.tables_in_both;
      merge work.base_tab_columns (in=Base) work.comp_tab_columns(in=Comp);
      by memname;
      if last.memname;
      if Base and Comp then
         output work.tables_in_both;
      else if Base and not Comp then
         output work.tables_not_in_comp;
      else if not Base and Comp then
         output work.new_tables_in_comp;
      keep memname;
   run;


   * Columns added/modified;
   data work.new_col_in_comp work.col_not_in_comp work.col_modified;
      merge work.base_tab_columns (in=base rename=(type=b_d_t length=b_d_l  format=b_d_f ))
            work.comp_tab_columns(in=comp rename=(type=c_d_t length=c_d_l  format=c_d_f ));
      by memname name;
      if base and comp then do;

      %if &include_fmt_in_diff=1 %then %do;
         if not (c_d_t = b_d_t & c_d_l = b_d_l &  b_d_f = c_d_f  ) then
            output work.col_modified;
         keep memname name c_d_t b_d_t c_d_l b_d_l  b_d_f c_d_f ;
      %end;
      %else %do  ;
         if not (c_d_t = b_d_t & c_d_l = b_d_l ) then
            output work.col_modified;
         keep memname name c_d_t b_d_t c_d_l b_d_l ;
      %end;

      end;
      else if not base and comp then
         output work.new_col_in_comp;
      else if base and not comp then
         output work.col_not_in_comp;

   run;

   data work.col_not_in_comp;
      merge work.col_not_in_comp work.tables_not_in_comp(in=tables) ;
      by memname;
      if not tables;
   run;

   data work.new_col_in_comp;
      merge work.new_col_in_comp work.new_tables_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   * Primary Key added/modified;
   data work.new_pk_in_comp work.pk_modified work.pk_not_in_comp;
      merge work.base_pk_columns (in=base  rename=(column_names=base_columns))
            work.comp_pk_columns (in=comp  rename=(column_names=comp_columns));
      by memname;
      if base and comp and not (base_columns = comp_columns) then
         output work.pk_modified;
      else if not base and comp then
         output work.new_pk_in_comp;
      else if base and not comp then
         output work.pk_not_in_comp;
   run;

   data work.pk_not_in_comp;
      merge work.pk_not_in_comp work.tables_not_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   data work.new_pk_in_comp;
      merge work.new_pk_in_comp work.new_tables_in_comp(in=tables);
      by memname;
      if not tables;
   run;


   * Not-Null constraints added/modified;
   data work.new_notnull_in_comp work.notnull_not_in_comp;
      merge work.base_not_null_cons (in=base)
            work.comp_not_null_cons (in=comp);
      by memname column_name;
      if not base and comp then
         output work.new_notnull_in_comp;
      else if base and not comp then
         output work.notnull_not_in_comp;
   run;

   data work.notnull_not_in_comp;
      merge work.notnull_not_in_comp work.tables_not_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   data work.new_notnull_in_comp;
      merge work.new_notnull_in_comp work.new_tables_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   * Indexes added/modified ;

   data work.new_ind_in_comp work.ind_modified work.ind_not_in_comp;
      merge work.base_index_columns (in=base rename=(name=base_columns unique_index=base_unique))
            work.comp_index_columns (in=comp rename=(name=comp_columns unique_index=comp_unique));
      by memname indxname;
      if base and comp and not (base_columns = comp_columns and
         base_unique = comp_unique) then
         output work.ind_modified;
      else if not base and comp then
         output work.new_ind_in_comp;
      else if base and not comp then
         output work.ind_not_in_comp;
   run;

   data work.ind_not_in_comp;
      merge work.ind_not_in_comp work.tables_not_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   data work.new_ind_in_comp;
      merge work.new_ind_in_comp work.new_tables_in_comp(in=tables);
      by memname;
      if not tables;
   run;

   * Write to difference report ;
   data _null_;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file;
      curr_time=put(datetime(),datetime25.);
      put 'DATABASE RECONCILE REPORT - ' curr_time ;
      put '------------------------------------------------' /;
      put 'BASELINE INFORMATION ' / '--------------------';
      put 'Version = ' "&Version";
      put 'Data Library = ' "&baseline";
      put 'Date = ' curr_time / ;
      put 'Production INFORMATION' / '------------------';
      put 'Data Library = ' "&compare";
      put 'Date = ' curr_time / ;
      put 'RECONCILE RESULTS' / '-----------------';

   run;

   data _null_;
      set work.tables_not_in_comp end=done;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then do;
         put '---Tables Not in ' "&compare";
      end;
      put '    ' _n_ memname;
      if done then put '---End Tables Not in ' "&compare" /;
   run;

   data _null_;
      set work.new_tables_in_comp end=done;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---New tables in ' "&compare" ;
      put '    ' _n_ memname;
      if done then put '---End New tables in ' "&compare" /;
   run;



   data _null_;
      set work.col_not_in_comp end=done;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Columns not in ' "&compare" ;
       x = cat(cats(memname,'.',name),' ',cats(b_d_t,'(',b_d_l,')'));
       put '    ' _n_ x;
      if done then put '---End Columns not in ' "&compare" /;
   run;

   data _null_;
      set work.col_modified end=done;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Columns modified in ' "&compare" ;

      if (b_d_t ne c_d_t) or (b_d_l ne c_d_l) then do;
         x=cat(cats(memname,'.',name),' ',b_d_t,'(',b_d_l,')');
         y=cat('-->',' ',c_d_t,'(',c_d_l,')');
         put '    ' _n_ x y;
      end;
      else if (b_d_f ne c_d_f ) and b_d_t=c_d_t and b_d_t='num'  then do;
         x = cat(cats(memname,'.',name),' ',cats(b_d_t,'(',b_d_l,')'),' ',b_d_f);
         y = cat('--> ',cats(c_d_t,'(',c_d_l,')'),' ',c_d_f);
         put '    ' _n_ x y;
      end;


      if done then put '---End Columns modified in ' "&compare" /;
   run;

   data _null_;
      set work.new_col_in_comp end=done; attrib y length=$140;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---New Columns in ' "&compare" ;
      y = cat(cats(memname,'.',name),' ',cats(c_d_t,'(',c_d_l,')'));
      put '    ' _n_ y;
      if done then put '---End New columns in ' "&compare" /;
   run;

   data _null_;
      set work.ind_not_in_comp end=done; attrib x length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Indexes not in ' "&compare" ;
      x = cats(memname,'.',indxname,'(',base_columns,')','*',base_unique);
      put '    ' _n_ x;
      if done then put '---End Indexes not in ' "&compare" /;
   run;

   data _null_;
      set work.ind_modified end=done; attrib x length=$400 y length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Indexes changed in ' "&compare" ;
      x = cats(memname,'.',indxname,'(',base_columns,')','*',base_unique);
      y = cats('-->','(',comp_columns,')','*',comp_unique);
      put '    ' _n_ x y;
      if done then put '---End Indexes changed in ' "&compare" /;
   run;

   data _null_;
      set work.New_ind_in_comp end=done; attrib y length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---New Indexes in ' "&compare" ;
      y = cats(memname,'.',indxname,'(',comp_columns,')','*',comp_unique);
      put '    ' _n_ y;
      if done then put '---End New indexes in ' "&compare" /;
   run;

   data _null_;
      set work.pk_not_in_comp end=done; attrib x length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Primary Keys not in ' "&compare" ;
      x = cats(memname,'(',base_columns,')');
      put '    ' _n_ x;
      if done then put '---End Primary Keys not in ' "&compare" /;
   run;

   data _null_;
      set work.pk_modified end=done; attrib x length=$400 y length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---Primary Keys modified in ' "&compare" ;
      x = cats(memname,'(',base_columns,')');
      y = cats('-->','(',comp_columns,')');
      put '    ' _n_ x y;
      if done then put '---End Primary Keys modified in ' "&compare" /;
   run;

   data _null_;
      set work.new_pk_in_comp end=done; attrib y length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---New Primary Keys in ' "&compare" ;
      y = cats(memname,'(',comp_columns,')');
      put '    ' _n_ y;
      if done then put '---End New Primary Keys in ' "&compare" /;
   run;

   data _null_;
      set work.notnull_not_in_comp end=done; attrib x length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---NotNull constraints missing in ' "&compare" ;
      x = cats(memname,'.',column_name);
      put '    ' _n_ x;
      if done then put '---End NotNull constraints missing in ' "&compare" /;
   run;

   data _null_;
      set work.new_notnull_in_comp end=done; attrib x length=$400;
      diff_file=cats("&reportpath","/schema_diff_&compare..txt");
      file compare filevar=diff_file mod;
      if _n_ eq 1 then put '---New NotNull constraints found in ' "&compare" ;
      x = cats(memname,'.',column_name);
      put '    ' _n_ x;
      if done then put '---End New NotNull constraints found in ' "&compare" /;
   run;


%mend compare_sas_lib;

/*

%sas_validate_schema(version= ,
                     reportpath=.
                    );

*/
